# Copyright (c) 2014, Warren Weckesser
# All rights reserved.
# See the LICENSE file for license information.

from numpy.testing import Tester as _Tester

from ._odeintw import odeintw


__version__ = "0.0.2.dev1"

test = _Tester().test



# Copyright (c) 2014, Warren Weckesser
# All rights reserved.
# See the LICENSE file for license information.

import numpy as np
from numpy.testing import (assert_allclose, assert_array_equal,
                           assert_, dec, run_module_suite)
import scipy
from scipy.integrate import odeint
from odeintw._odeintw import (_complex_to_real_jac, _transform_banded_jac,
                              odeintw)


_scipy_version = tuple(int(n) for n in scipy.__version__.split('.')[:3])
_banded_bug_msg = "known bug in scipy.integrate.odeint for scipy versions before 0.14.0"


def test_complex_to_real_jac():
    z = np.array([[1+2j]])
    r = _complex_to_real_jac(z)
    yield assert_array_equal, r, np.array([[1, -2], [2, 1]])
    z = np.array([[1+2j, 3+4j],
                  [5+6j, 7+8j]])
    r = _complex_to_real_jac(z)
    expected = np.array([[1, -2, 3, -4],
                         [2,  1, 4,  3],
                         [5, -6, 7, -8],
                         [6,  5, 8,  7]])
    yield assert_array_equal, r, expected


@dec.knownfailureif(_scipy_version < (0, 14, 0), _banded_bug_msg) 
def test_transform_banded_jac():
    j = np.array([[0,  0,  1,  2],
                  [0,  0,  3,  4],
                  [5,  6,  7,  8],
                  [9, 10, 11, 12]])
    t = _transform_banded_jac(j)
    expected = np.array([[0,  0,  0,  2],
                         [0,  0,  1,  4],
                         [0,  6,  3,  8],
                         [5, 10,  7, 12],
                         [9,  0, 11,  0]])
    yield assert_array_equal, t, expected


def system1_complex(z, t):
    return z * (1 - z)

def system1_real(z, t):
    x, y = z
    return [x*(1-x) + y**2, y - 2*x*y]


def test_odeintw_complex():
    z0 = 0.25 + 0.5j
    t = np.linspace(0, 1, 5)
    zsol = odeintw(system1_complex, z0, t)
    zv = zsol.view(np.float64)
    sol = odeintw(system1_real, [z0.real, z0.imag], t)
    yield assert_allclose, zv, sol


def system2_array(a, t):
    return -a.T

def system2_vector(a, t):
    a11, a12, a21, a22 = a
    return [-a11, -a21, -a12, -a22]

def test_odeint_array():
    t = np.linspace(0, 1, 5)
    a0 = np.array([[1, -2], [3, -4]])
    asol = odeintw(system2_array, a0, t)
    avec = asol.reshape(-1, 4)
    sol = odeintw(system2_vector, a0.ravel(), t)
    yield assert_allclose, avec, sol


def system3_func(y, t, c):
    return c.dot(y)


def system3_funcz(y, t, c):
    # Same calculation as `system3_func`, but computed using real arrays,
    # so the calculation in `dot` should follow the same code path for
    # both the real and complex examples below.
    creal = _complex_to_real_jac(c)
    dydt = creal.dot(y.view(np.float64))
    return dydt.view(np.complex128)


def system3_jac(y, t, c):
    return c

def system3_bjac_cols(y, t, c):
    return np.column_stack( (np.r_[0, np.diag(c, 1)], np.diag(c)) )

def system3_bjac_rows(y, t, c):
    return np.row_stack( (np.r_[0, np.diag(c, 1)], np.diag(c)) )


def test_system3():
    c = np.array([[-20+1j, 5-1j,      0,       0], 
                  [     0, -0.1,  1+2.5j,      0],
                  [     0,    0,      -1,    0.5],
                  [     0,    0,       0,  -5+10j]])

    z0 = np.arange(1,5.0) + 0.5j

    t = np.linspace(0, 250, 11)

    common_kwargs = dict(full_output=True, atol=1e-12, rtol=1e-10,
                         mxstep=1000)

    sol0, info0 = odeintw(system3_funcz, z0, t, Dfun=system3_jac,
                          args=(c,), **common_kwargs)
    nje0 = info0['nje']

    x0 = z0.view(np.float64)
    sol1, info1 = odeint(system3_func, x0, t, Dfun=system3_jac,
                         args=(_complex_to_real_jac(c),), **common_kwargs)
    nje1 = info1['nje']

    # Using assert_array_equal here is risky.  The system definitions have
    # been defined so the call to odeint in odeintw follows the same
    # code path as the call to odeint above.  Still, floating point operations
    # aren't necessarily deterministic.
    yield assert_array_equal, sol0.view(np.float64), sol1
    yield assert_array_equal, nje0, nje1


@dec.knownfailureif(_scipy_version < (0, 14, 0), _banded_bug_msg) 
def test_system3_banded():
    c = np.array([[-20+1j, 5-1j,      0,       0], 
                  [     0, -0.1,  1+2.5j,      0],
                  [     0,    0,      -1,    0.5],
                  [     0,    0,       0,  -5+10j]])

    common_kwargs = dict(args=(c,), full_output=True, atol=1e-12, rtol=1e-10,
                         mxstep=1000)
    z0 = np.arange(1,5.0) + 0.5j
    t = np.linspace(0, 250, 11)

    sol0, info0 = odeintw(system3_func, z0, t, Dfun=system3_jac,
                          **common_kwargs)

    sol1, info1 = odeintw(system3_func, z0, t, Dfun=system3_bjac_cols,
                          ml=0, mu=1, col_deriv=True, **common_kwargs)
    sol2, info2 = odeintw(system3_func, z0, t, Dfun=system3_bjac_rows,
                          ml=0, mu=1, **common_kwargs)

    yield assert_allclose, sol0, sol1
    yield assert_allclose, sol0, sol2
    # The same code paths should have been followed in computing
    # sol1 and sol2, so the number of jacobian evaluations should
    # be the same.
    yield assert_array_equal, info1['nje'], info2['nje']


if __name__ == "__main__":
    run_module_suite()



# non-resonant leptogenesis with one decaying sterile neutrino using the Boltzmann equations. Note these kinetic equations do not include off diagonal flavour oscillations. Equations from 1112.4528
import ulysses
import numpy as np
from odeintw import odeintw


from numba import jit
@jit
def fast_RHS(y0, d, w1, n1eq, epstt,epsmm,epsee,c1t,c1m,c1e):
    N1      = y0[0]
    Ntt     = y0[1]
    Nmm     = y0[2]
    Nee     = y0[3]
    c1tc    = np.conjugate(c1t)
    c1mc    = np.conjugate(c1m)
    c1ec    = np.conjugate(c1e)
    rhs1 =      -d*(N1-n1eq)

    rhs2 = epstt*d*(N1-n1eq)-0.5*w1*(2*c1t*c1tc*Ntt)
    rhs3 = epsmm*d*(N1-n1eq)-0.5*w1*(2*c1m*c1mc*Nmm)
    rhs4 = epsee*d*(N1-n1eq)-0.5*w1*(2*c1e*c1ec*Nee)

    return [rhs1, rhs2, rhs3, rhs4]

class EtaB_1BE(ulysses.ULSBase):
    """
    Boltzmann equation (BE) with one decaying sterile. See arxiv:1112.4528.
    Note these kinetic equations do not include off diagonal flavour
    oscillations.
    """

    def RHS(self, y0,z,epstt,epsmm,epsee,c1t,c1m,c1e,k):

        if z != self._currz or z == self.zmin:
            self._d       = np.real(self.D1(k,z))
            self._w1      = np.real(self.W1(k,z))
            self._n1eq    = self.N1Eq(z)
            self._currz=z

        return fast_RHS(y0, self._d, self._w1, self._n1eq, epstt,epsmm,epsee,c1t,c1m,c1e)

    @property
    def EtaB(self):
        #Define fixed quantities for BEs
        epstt = np.real(self.epsilon1ab(2,2))
        epsmm = np.real(self.epsilon1ab(1,1))
        epsee = np.real(self.epsilon1ab(0,0))

        c1t   =                 self.c1a(2)
        c1m   =                 self.c1a(1)
        c1e   =                 self.c1a(0)

        k       = np.real(self.k1)
        y0      = np.array([0+0j,0+0j,0+0j,0+0j], dtype=np.complex128)

        params  = np.array([epstt,epsmm,epsee,c1t,c1m,c1e,k], dtype=np.complex128)

        ys      = odeintw(self.RHS, y0, self.zs, args = tuple(params))
        nb      = self.sphalfact*(ys[-1,1]+ys[-1,2]+ys[-1,3])

        self.ys  = np.real(ys[:, [1,2,3]])

        return np.real(nb)



import cmath
import ulysses

def readConfig(fname):
    from collections import OrderedDict

    fixed  = OrderedDict()
    ranges = OrderedDict()
    with open(fname) as f:
        for line in f:
            l=line.strip()
            if l.startswith("#"):
                continue
            if len(l) == 0:
                continue

            fields = l.split()
            if len(fields)==1:
                print( "Warning, not understood instruction:", l)
                continue

            elif len(fields)==2:
                try:
                    fixed[fields[0]] = float(fields[1])
                except:
                    fixed[fields[0]] = complex(fields[1])

            elif len(fields)==3:
                ranges[fields[0]] = (float(fields[1]), float(fields[2]))

            else:
                print ("Warning, not understood instruction:", l)
                continue
        return ranges, fixed

def selectModel(model, **kwargs):
    r"""

    This function loads and returns an instance of one of the built-in models.
    The kwargs are passed on to the base class.

        :Arguments:
            * *model* (``str``) --
              The model. Must be in ["1DME", "2DME", "3DME", "1BE", "2BE", "2resonant", "3DMEsct", "3DMErhtau"]

    """
    import ulysses
    if not ":" in model:
        avail = ["1DME", "2DME", "3DME", "1BE", "2BE", "2resonant", "3DMEsct", "3DMErhtau"]
        if   model=="1DME":                    return ulysses.EtaB_1DME(**kwargs)
        elif model=="2DME":                    return ulysses.EtaB_2DME(**kwargs)
        elif model=="3DME":                    return ulysses.EtaB_3DME(**kwargs)
        elif model=="1BE":                     return ulysses.EtaB_1BE(**kwargs)
        elif model=="2BE":                     return ulysses.EtaB_2BE(**kwargs)
        elif model=="2resonant":               return ulysses.EtaB_2Resonant(**kwargs)
        elif model=="3DMEsct":                 return ulysses.EtaB_3DME_Scattering(**kwargs)
        elif model=="3DMErhtau":               return ulysses.EtaB_3DS_Scattering_RHtaur(**kwargs)
        else:
            raise Exception("Specified model '{}' unknown.\n Select from: {}".format(model, avail))
    else:
        return loadPlugin(model, **kwargs)

def loadPlugin(model, **kwargs):
    r"""
    Plugin loader. This attemts to load and return and instance of
    the class CLASS in the file FILENAME. Both are given as a single
    string using : as separator. The kwargs are passed on to the base class.

        :Arguments:
            * *model* (``str``) --
              FILENAME:CLASS

    """
    import ulysses
    m_file, m_name = model.split(":")
    import os
    if not os.path.exists(m_file):
        raise Exception("Specified module file '{}' does not exist.{}".format(m_file))

    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    from importlib.machinery import SourceFileLoader

    foo = SourceFileLoader(m_name, m_file).load_module()

    # https://stackoverflow.com/questions/1796180/how-can-i-get-a-list-of-all-classes-within-current-module-in-python
    import inspect

    IM = dict(inspect.getmembers(foo, inspect.isclass))
    if not m_name in IM:
        raise Exception("Specified class '{}'  not found in module {}".format(m_name, m_file))
    if not issubclass(IM[m_name], ulysses.ULSBase):
        raise Exception("Specified class '{}'  does not derive from leptoms.ULSBase".format(m_name))

    return IM[m_name](**kwargs)



from scipy.special import kn
import numpy as np
from odeintw import odeintw
import timeit
from numpy import linalg

def fast_isPerturbative(y):
    """
    Check perturbativity of Yukawas
    """
    #limit of perturbativity for y
    limit = np.power(4*np.pi,0.5)
    #check if any element of column 1, 2 or 3 is larger than limit
    col1               = (y[0,0] < limit)*(y[1,0] < limit)*(y[2,0] < limit)
    col2               = (y[0,1] < limit)*(y[1,1] < limit)*(y[2,1] < limit)
    col3               = (y[0,2] < limit)*(y[1,2] < limit)*(y[2,2] < limit)
    return col1*col2*col3

def my_kn1(x):
    """
    Convenience wrapper for kn(1, x)
    """
    return kn(1, x) if x<=600 else 1e-100#3. + 8.*x

def my_kn2(x):
    """
    Convenience wrapper for kn(2, x)
    """
    return kn(2, x) if x<=600 else 1e-100#15. + 8.*x



# This is the base class
class ULSBase(object):
    def __init__(self, *args, **kwargs):
        r"""
        Base class for all EtaB calculators.

        Set global constants here.

        :Keyword Arguments:
            * *vev* (``float``) --
              Higgs VEV in GeV
            * *mhiggs* (``float``) --
              Higgs mass in GeV
            * *mz* (``float``) --
              Z-boson mass in GeV
            * *gstar* (``float``) --
              Relativistic degrees of freedom at high temperature
            * *mplanck* (``float``) --
              Planck mass in GeV
            * *mstar* (``float``) --
              Neutrino cosmological mass in GeV


        """
        #Higgs vev, mass and Z-mass in GeV
        self.v =  kwargs["vev"]                        if kwargs.get("vev")      is not None else 174.
        self.MH = kwargs["mhiggs"]                     if kwargs.get("mhiggs")   is not None else 125.
        self.MZ = kwargs["mz"]                         if kwargs.get("mz")       is not None else 91.1876
        #relativistic degrees of freedom at high temperature
        self.gstar = kwargs["gstar"]                   if kwargs.get("gstar")    is not None else 106.75
        #Planck mass in GeV
        self.MP    = kwargs["mplanck"]                 if kwargs.get("mplanck")  is not None else 1.22e+19
        #neutrino cosmological mass in GeV
        self.mstar = kwargs["mstar"]                   if kwargs.get("mstar")    is not None else 1.0e-12
        # Mass-splittings, all in GeV^2
        self.msplit2_solar       = kwargs["m2solar"]   if kwargs.get("m2solar")  is not None else 7.390e-5*1e-18 # 2018
        self.msplit2_athm_normal = kwargs["m2atm"]     if kwargs.get("m2atm")    is not None else 2.525e-3*1e-18 # Values
        self.msplit2_athm_invert = kwargs["m2atminv"]  if kwargs.get("m2atminv") is not None else 2.512e-3*1e-18 # from nu-fit 4.0

        # Flags
        self.debug   = kwargs["debug"]                 if kwargs.get("debug")    is not None else False

        # Parameters of the solver
        self._zmin   = kwargs["zmin"]                  if kwargs.get("zmin")     is not None else 0.1
        self._zmax   = kwargs["zmax"]                  if kwargs.get("zmax")     is not None else 1000
        self._zsteps = kwargs["zsteps"]                if kwargs.get("zsteps")   is not None else 1000
        self._currz = self.zmin

        # Model switches
        self.ordering = kwargs["ordering"]             if kwargs.get("ordering") is not None else 0
        self.loop     = kwargs["loop"]                 if kwargs.get("loop")     is not None else False

        self.zs=None
        self.ys=None
        self.setZS()
        self.sphalfact = kwargs["sphalfact"] if kwargs.get("sphalfact") is not None else 0.01

    @property
    def constants(self):
        s=" Global constants:"
        s+= "\n\t Higgs VEV {} GeV ['vev']".format(self.v)
        s+= "\n\t Higgs mass {} GeV ['mhiggs']".format(self.MH)
        s+= "\n\t Z-boson mass {} GeV ['mz']".format(self.MZ)
        s+= "\n\t Planck mass {} GeV ['mplanck']".format(self.MP)
        s+= "\n\t Neutrino cosmological mass {} GeV ['mstar']".format(self.mstar)
        s+= "\n\t Relativistic degrees of freedom at high temperature {} GeV ['gstar']".format(self.gstar)
        s+= "\n\t Solar mass square splitting {} GeV^2 ['m2solar']".format(self.msplit2_solar)
        s+= "\n\t Atmospheric mass square splitting, normal ordering {} GeV^2 ['m2atm']".format(self.msplit2_athm_normal)
        s+= "\n\t Atmospheric mass square splitting, inverted ordering {} GeV^2 ['m2atminv']".format(self.msplit2_athm_invert)
        return s

    def __call__(self, x):
        r"""
        Operator that returns EtaB for a given parameter point.

        :Arguments:
            * *x* (``dict``) --
              parameter dictionary

        NOTE --- this operator is intended to be used with derived classes where EtaB is implemented
        """
        self.setParams(x)
        return self.EtaB

    def __str__(self):
        s="Model:\n{}".format(self.__doc__)
        s+= "\nNormal ordering\n" if self.ordering==0 else "\nInverted ordering\n"
        s+= "Loop-corrected Yukawa\n" if self.loop else "Tree-level Yukawa\n"
        s+="Integration in [{}, {}] in {} steps\n".format(self._zmin, self._zmax, self._zsteps)
        if self.debug: s+=self.constants
        return s

    def setZMin(self, x):
        self._zmin=x
        self.setZS()

    def setZMax(self, x):
        self._zmax=x
        self.setZS()

    def setZSteps(self, x):
        self._zsteps=x
        self.setZS()

    def setZS(self):
        self.zs = np.geomspace(self.zmin, self.zmax, self.zsteps)
        if self.debug:
            print("Integration range:",self.zs.min(),self.zs.max())

    @property
    def zmin(self): return self._zmin

    @property
    def zmax(self): return self._zmax

    @property
    def zsteps(self): return self._zsteps

    @property
    def evolData(self):
        r"""

        :getter: Return a 4-D array of the evolution data.

        The first column is the evolution variable
        The second column corresponds to Ntautau, the
        third to Nmumu and the last columnd to Nee

        """
        pd = np.empty((self.zsteps, 4))
        pd[:,      0] = self.zs
        pd[:,[1,2,3]] = self.ys
        return pd

    def setParams(self, pdict):
        """
        This set the model parameters. pdict is expected to be a dictionary
        """
        self.delta    = pdict['delta']/180*np.pi
        self.a        = pdict['a21']/180*np.pi
        self.b        = pdict['a31']/180*np.pi
        self.theta12  = pdict['t12']/180*np.pi
        self.theta23  = pdict['t23']/180*np.pi
        self.theta13  = pdict['t13']/180*np.pi
        self.x1       = pdict['x1']/180*np.pi
        self.y1       = pdict['y1']/180*np.pi
        self.x2       = pdict['x2']/180*np.pi
        self.y2       = pdict['y2']/180*np.pi
        self.x3       = pdict['x3']/180*np.pi
        self.y3       = pdict['y3']/180*np.pi
        self.m1       = 10**pdict['m'] * 1e-9 # NOTE input is in log10(m1) in eV --- we convert here to the real value in GeV
        self.M1       = 10**pdict['M1']  #
        self.M2       = 10**pdict['M2']  #
        self.M3       = 10**pdict['M3']  #


    def printParams(self):
        """
        Print current parameters.
        """
        K = ('delta','a21','a31','t12','t23','t13','x1','y1','x2','y2','x3','y3','m','M1','M2','M3')
        V = (self.delta/np.pi*180, self.a/np.pi*180, self.b/np.pi*180, self.theta12/np.pi*180, self.theta23/np.pi*180, self.theta13/np.pi*180, self.x1/np.pi*180, self.y1/np.pi*180, self.x2/np.pi*180, self.y2/np.pi*180, self.x3/np.pi*180, self.y3/np.pi*180,
                np.log10(self.m1/1e-9), np.log10(self.M1), np.log10(self.M2), np.log10(self.M3))
        for k, v in zip(K,V):
            print(k,v)


    # Some general calculators purely based on input parameters
    @property
    def R(self):
        """
        Orthogonal matrix R = R1.R2.R3
        """

        R1 = np.array([[1., 0., 0.],
                       [0.,  np.cos(self.x1+self.y1*1j), np.sin(self.x1+self.y1*1j)],
                       [0., -np.sin(self.x1+self.y1*1j), np.cos(self.x1+self.y1*1j)]], dtype=np.complex128)

        R2 = np.array([[ np.cos(self.x2+self.y2*1j), 0., np.sin(self.x2+self.y2*1j)],
                       [0., 1. , 0.],
                       [-np.sin(self.x2+self.y2*1j), 0., np.cos(self.x2+self.y2*1j)]], dtype=np.complex128)

        R3 = np.array([[ np.cos(self.x3+self.y3*1j), np.sin(self.x3+self.y3*1j), 0.],
                       [-np.sin(self.x3+self.y3*1j), np.cos(self.x3+self.y3*1j), 0.],
                       [0., 0., 1.]], dtype=np.complex128)

        return R1 @ R2 @ R3

    @property
    def SqrtDM(self):
        """
        Square root of diagonal heavy mass matrix.
        TODO: is this np.sqrt(self.DM) ???
        """
        return np.array([[np.sqrt(self.M1), 0., 0.],
                         [0., np.sqrt(self.M2), 0.],
                         [0., 0., np.sqrt(self.M3)]], dtype=np.complex128)

    @property
    def DM(self):
        """
        Diagonal heavy mass matrix.
        """
        return np.array([[self.M1, 0., 0.],
                         [0., self.M2, 0.],
                         [0., 0., self.M3]], dtype=np.complex128)

    @property
    def SqrtDm(self):
        """
        Square root of diagonal light mass matrix.
        Everything is in GeV.
        """

        if self.ordering==0:
            m11 = np.sqrt(self.m1)
            m22 = np.sqrt(np.sqrt(self.msplit2_solar       + self.m1*self.m1))
            m33 = np.sqrt(np.sqrt(self.msplit2_athm_normal + self.m1*self.m1))
        elif self.ordering==1:
            m11 = np.sqrt(np.sqrt(self.msplit2_athm_invert + self.m1*self.m1 - self.msplit2_solar))
            m22 = np.sqrt(np.sqrt(self.msplit2_athm_invert + self.m1*self.m1))
            m33 = np.sqrt(self.m1)
        else:
            raise Exception("ordering %i not implemented"%self.ordering)

        return np.array([ [m11,  0.,  0.],
                          [ 0., m22,  0.],
                          [ 0.,  0., m33] ], dtype=np.complex128)

    @property
    def U(self):
        """
        PMNS matrix using the PDG parametrisation convention.
        """
        s12     = np.sin(self.theta12)
        s23     = np.sin(self.theta23)
        s13     = np.sin(self.theta13)
        c12     = np.power(1-s12*s12,0.5)
        c23     = np.power(1-s23*s23,0.5)
        c13     = np.power(1-s13*s13,0.5)
        return np.array([ [c12*c13,c13*s12*np.exp(self.a*1j/2.), s13*np.exp(self.b*1j/2-self.delta*1j)],
                           [-c23*s12 - c12*np.exp(self.delta*1j)*s13*s23,np.exp((self.a*1j)/2.)*(c12*c23 - np.exp(self.delta*1j)*s12*s13*s23) , c13*np.exp((self.b*1j)/2.)*s23],
                           [-c12*c23*np.exp(self.delta*1j)*s13 + s12*s23,np.exp((self.a*1j)/2.)*(-c23*np.exp(self.delta*1j)*s12*s13 - c12*s23) ,c13*c23*np.exp((self.b*1j)/2.)]], dtype=np.complex128)


    @property
    def fMR(self):
        """
        This function returns the diagonl heavy neutrino mass matrix taking
        radiative corrections into account. (see h_loop) I.e. equivalent to self.SqrtDM in loop case
        """
        a = 1./self.M1
        b = 1./self.M2
        c = 1./self.M3

        prefactor = -1./(32*np.pi**2*self.v**2)
        d = self.fMLoop(self.M1)
        e = self.fMLoop(self.M2)
        f = self.fMLoop(self.M3)
        A = np.diag([a,b,c])
        B = prefactor*np.diag([d,e,f])

        return np.sqrt(linalg.inv(A+B))


    @property
    def fMLoopHelper(self):
        """
        Helper function.
        """
        prefactor = 1./(32*np.pi**2*self.v**2)
        d = self.fMLoop(self.M1)
        e = self.fMLoop(self.M2)
        f = self.fMLoop(self.M3)
        B = prefactor*np.diag([d,e,f])
        return B

    def fMLoop(self, x):
        """
        The loop function.
        """
        rH2 = (x/self.MH)**2
        rZ2 = (x/self.MZ)**2
        return x*(np.log(rH2)/(rH2-1) + 3*np.log(rZ2)/(rZ2-1) )

    @property
    def m_tree(self):
        """
        Tree-level mass matrix.
        """
        return self.v**2 * self.h @ np.linalg.inv(self.DM) @ np.transpose(self.h)

    @property
    def m_loop(self):
        """
        One-loop-level mass matrix.
        """
        return -1 * self.v**2 * self.h @ self.fMLoopHelper @ np.transpose(self.h)

    @property
    def h(self):
        """
        YUKAWA matrix.
        """
        return self.h_loop if self.loop else self.h_tree

    @property
    def h_loop(self):
        """
        Yukawa matrix (LOOP + Tree).
        """
        return (1./self.v)*(self.U @ self.SqrtDm @ np.transpose(self.R) @ self.fMR)

    @property
    def h_tree(self):
        """
        Yukawa matrix, tree-level.
        """
        return (1./self.v)*(self.U @ self.SqrtDm @ np.transpose(self.R) @ self.SqrtDM)

    @property
    def meff1(self):
        """
        Effective mass 1 used for decay ans washout.
        """
        return np.dot(np.conjugate(np.transpose(self.h)),self.h)[0,0]*(self.v**2)/self.M1

    @property
    def meff2(self,):
        """
        Effective mass 2 used for decay ans washout.
        """
        return np.dot(np.conjugate(np.transpose(self.h)),self.h)[1,1]*(self.v**2)/self.M2

    @property
    def meff3(self,):
        """
        Effective mass 3 used for decay and washout.
        """
        return np.dot(np.conjugate(np.transpose(self.h)),self.h)[2,2]*(self.v**2)/self.M3

    @property
    def k1(self):
        """
        Decay parameter 1
        """
        return self.meff1/self.mstar

    @property
    def k2(self):
        """
        Decay parameter 2
        """
        return self.meff2/self.mstar

    @property
    def k3(self):
        """
        Decay parameter 3
        """
        return self.meff3/self.mstar

    def D1(self, k1, z):
        """
        Decay term for Boltzmann equation
        """
        return  k1*z*my_kn1( z)/my_kn2( z)

    def scat(self, z):
        """
        Function that multiplies washouts to incorporate scatterings.
        """
        M         = self.DM
        prefactor = 0.1*(1+15/(8*z))
        t         = np.log(M[0,0]/self.MH)
        bracket   = z*t*np.log(1+(1/z)*8.77298/t)+1/z
        return prefactor*bracket

    def DS(self, k1, z):
        """
        Function that replaces decay term when scattering is also present
        """
        M         = self.DM
        t         = np.log(M[0,0]/self.MH)
        DStemp    = 0.1*k1*(1+t*(z**2)*np.log(1+8.77298/(t*z)))
        return DStemp

    def D2(self, k,z):
        """
        Decay term for Boltzmann equation with two decaying steriles.
        """
        r =self.M2/self.M1
        a = r*r
        x=np.real(r*z)
        b = my_kn1(x)
        c = my_kn2(x)
        return k*z*a*b/c

    def D3(self, k,z):
        """
        Decay term for Boltzmann equation with three decaying steriles.
        """
        r =self.M3/self.M1
        a = r*r
        x=np.real(r*z)
        b = my_kn1(x)
        c = my_kn2(x)
        return k*z*a*b/c

    def N1Eq(self, z):
        """
        Equilibrium number density with one decaying sterile.
        """
        n1 = 3./8.*(z**2)*my_kn2(z)
        return n1

    def N2Eq(self, z):
        """
        Equilibrium number density with two decaying steriles.
        """
        r = self.M2/self.M1
        n2 = 3./8.*np.power(r*z,2)*my_kn2(r*z)
        return n2

    def N3Eq(self, z):
        """
        Equilibrium number density with three decaying steriles.
        """
        r = self.M3/self.M1
        n3 = 3./8.*np.power(r*z,2)*my_kn2(r*z)
        return n3

    def W1(self, k1, z):
        """
        Washout parameter with one decaying sterile.
        """
        w1 = 1./4*(z**3)*k1*my_kn1(z)
        return w1

    def W2(self, k, z):
        """
        Washout parameter with two decaying steriles.
        """
        r = self.M2/self.M1
        w2 = k*r/4*np.power(r*z,3) * my_kn1(r*z)
        return w2

    def W3(self, k, z):
        """
        Washout parameter with three decaying steriles.
        """
        r = self.M3/self.M1
        w3 = k*r/4*np.power(r*z,3) * my_kn1(r*z)
        return w3

    def hterm(self, a, b):
        """
        Projection probability projecting onto certain directions
        of flavour space indicated by the indices a and b.

        a ... [0,1,2]
              0 = e
              1 = mu
              2 = tau
        b ... [0,1]
              0 = term1
              1 = term2

        """
        norm          = 1./((np.dot(np.conjugate(np.transpose(self.h)), self.h))[0,0])
        return norm*(np.abs(self.h[a,b])**2)

    def c1a(self, a):
        """
        Probability coefficient for 1 a
        """
        norm          = np.sqrt(1./((np.dot(np.conjugate(np.transpose(self.h)), self.h))[0,0]))
        return norm*(self.h[a,0])

    def c2a(self, a):
        """
        Probability coefficient for 2 a
        """
        norm          = np.sqrt(1./((np.dot(np.conjugate(np.transpose(self.h)), self.h))[1,1]))
        return norm*(self.h[a,1])

    def c3a(self, a):
        """
        Probability coefficient for 3 a
        """
        norm          = np.sqrt(1./((np.dot(np.conjugate(np.transpose(self.h)), self.h))[2,2]))
        return norm*(self.h[a,2])

    def f1(self, x):
        """
        Loop function in epsilon.
        """
        r2=np.power(x,2)

        f1temp = 2./3.*r2*( (1.+r2) * np.log( (1.+r2) / r2 ) - (2.-r2)/(1.-r2) )

        return f1temp if x<10000 else 1

    def f2(self, x):
        """
        Loop function in epsilon.
        """
        return (2./3.)*(1/(np.power(x,2)-1.))

    def epsilon(self, i, j, k, m):
        """
        CP asymmetry parameter.
        i,j,k,m denote indices in the heavy neutrino mass matrix.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        #define terms of epsilon: prefactor and first term (first), second term (second) etc.
        prefactor   = (3/(16*np.pi))*(1/(lsquare[i,i]))
        first       = np.imag(lsquare[i,j]*l[m,j]*lcon[m,i])*(M[i,i]/M[j,j])*self.f1(M[j,j]/M[i,i])

        second      = np.imag(lsquare[j,i]*l[m,j]*lcon[m,i])*(2./3.)*(1/(np.power(M[j,j]/M[i,i],2)-1.))
        third       = np.imag(lsquare[i,k]*l[m,k]*lcon[m,i])*(M[i,i]/M[k,k])*self.f1(M[k,k]/M[i,i])
        fourth      = np.imag(lsquare[k,i]*l[m,k]*lcon[m,i])*(2./3.)*(1/(np.power(M[k,k]/M[i,i],2)-1.))
        return prefactor*(first+second+third+fourth)

    def epsilon1ab(self,a,b):
        """
        Off-diagonal CP asymmetry parameter for decays of N1. a and b denote lepton flavour.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        #define terms of epsilon: prefactor and first term (first), second term (second) etc.
        prefactor   = (3/(32*np.pi))*(1/(lsquare[0,0]))
        first       = 1j*(lsquare[1,0]*l[a,0]*lcon[b,1]-lsquare[0,1]*l[a,1]*lcon[b,0])*(M[0,0]/M[1,1])*self.f1(M[1,1]/M[0,0])
        third       = 1j*(lsquare[2,0]*l[a,0]*lcon[b,2]-lsquare[0,2]*l[a,2]*lcon[b,0])*(M[0,0]/M[2,2])*self.f1(M[2,2]/M[0,0])
        second      = 1j*(2./3.)*(1/(np.power(M[1,1]/M[0,0],2)-1.))*(l[a,0]*lcon[b,1]*lsquare[0,1]-lcon[b,0]*l[a,1]*lsquare[1,0])
        fourth      = 1j*(2./3.)*(1/(np.power(M[2,2]/M[0,0],2)-1.))*(l[a,0]*lcon[b,2]*lsquare[0,2]-lcon[b,0]*l[a,2]*lsquare[2,0])
        epsilon1abtemp = prefactor*(first+second+third+fourth)
        return epsilon1abtemp

    def epsilon2ab(self,a,b):
        """
        Off-diagonal CP asymmetry parameter for decays of N2. a and b denote lepton flavour.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        #define terms of epsilon: prefactor and first term (first), second term (second) etc.
        prefactor   = (3/(32*np.pi))*(1/(lsquare[1,1]))
        first       = 1j*(lsquare[0,1]*l[a,1]*lcon[b,0]-lsquare[1,0]*l[a,0]*lcon[b,1])*(M[1,1]/M[0,0])*self.f1(M[0,0]/M[1,1])
        third       = 1j*(lsquare[2,1]*l[a,1]*lcon[b,2]-lsquare[1,2]*l[a,2]*lcon[b,1])*(M[1,1]/M[2,2])*self.f1(M[2,2]/M[1,1])
        second      = 1j*(2./3.)*(1/(np.power(M[0,0]/M[1,1],2)-1.))*(l[a,1]*lcon[b,0]*lsquare[1,0]-lcon[b,1]*l[a,0]*lsquare[0,1])
        fourth      = 1j*(2./3.)*(1/(np.power(M[2,2]/M[1,1],2)-1.))*(l[a,1]*lcon[b,2]*lsquare[1,2]-lcon[b,1]*l[a,2]*lsquare[2,1])
        epsilon2abtemp = prefactor*(first+second+third+fourth)
        return epsilon2abtemp

    def epsilon3ab(self,a,b):
        """
        Off-diagonal CP asymmetry parameter for decays of N3. a and b denote lepton flavour.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        #define terms of epsilon: prefactor and first term (first), second term (second) etc.
        prefactor   = (3/(32*np.pi))*(1/(lsquare[2,2]))
        first       = 1j*(lsquare[0,2]*l[a,2]*lcon[b,0]-lsquare[2,0]*l[a,0]*lcon[b,2])*(M[2,2]/M[0,0])*self.f1(M[0,0]/M[2,2])
        third       = 1j*(lsquare[1,2]*l[a,2]*lcon[b,1]-lsquare[2,1]*l[a,1]*lcon[b,2])*(M[2,2]/M[1,1])*self.f1(M[1,1]/M[2,2])
        second      = 1j*(2./3.)*(1/(np.power(M[0,0]/M[2,2],2)-1.))*(l[a,2]*lcon[b,0]*lsquare[2,0]-lcon[b,2]*l[a,0]*lsquare[0,2])
        fourth      = 1j*(2./3.)*(1/(np.power(M[1,1]/M[2,2],2)-1.))*(l[a,2]*lcon[b,1]*lsquare[2,1]-lcon[b,2]*l[a,1]*lsquare[1,2])
        epsilon3abtemp = prefactor*(first+second+third+fourth)
        return epsilon3abtemp

    @property
    def eps100(self): return self.epsilon1ab(0,0)

    @property
    def eps111(self): return self.epsilon1ab(1,1)

    @property
    def eps122(self): return self.epsilon1ab(2,2)

    @property
    def eps200(self): return self.epsilon2ab(0,0)

    @property
    def eps211(self): return self.epsilon2ab(1,1)

    @property
    def eps222(self): return self.epsilon2ab(2,2)

    @property
    def eps300(self): return self.epsilon3ab(0,0)

    @property
    def eps311(self): return self.epsilon3ab(1,1)

    @property
    def eps322(self): return self.epsilon3ab(2,2)

    def epsilonaaRES(self,a):
        """
        CP asymmetry for resonant Leptogenesis.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)
        gamma1    = (lsquare[0,0]/(8*np.pi))*M[0,0]
        gamma2    = (lsquare[1,1]/(8*np.pi))*M[1,1]
        DeltaM    = M[1,1]-M[0,0]
        sum1      = np.imag(ldag[0,a]*l[a,1]*lsquare[0,1])+np.imag(ldag[0,a]*l[a,1]*lsquare[1,0])
        epsbar1   = sum1/(lsquare[0,0]*lsquare[1,1])
        fmix      =  -2*(DeltaM/gamma2)/(1+(2*DeltaM/gamma2)**2)
        fosc      = -0.5*(DeltaM/gamma2)/(1+(DeltaM/gamma2)**2)
        epsbar    = -0.5*epsbar1 * (fosc + fmix)
        return epsbar

    def resonance(self, z):
        """
        calculate decay rate Gamma in terms of the total epsilon (epstot)
        """
        eps1tau = np.real(self.epsilon(0,1,2,2))
        eps1mu  = np.real(self.epsilon(0,1,2,1))
        eps1e   = np.real(self.epsilon(0,1,2,0))
        epstot  = eps1tau+eps1mu+eps1e
        k       = self.k1
        d       = self.D1(k,z)
        Gamma   = (self.M1**2/self.MP)*np.sqrt(2*np.pi/3)*np.sqrt((np.pi**2)*self.gstar/30)*(1+epstot)*d/z

        #calculate (decay rate)/(mass splitting)
        return Gamma/(M2-M1)

    @property
    def Gamma1(self):
        """
        Decay rate of N1.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        return (M[0,0]/(8*np.pi))*lsquare[0,0]

    @property
    def Gamma2(self):
        """
        Decay rate of N2.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        return (M[1,1]/(8*np.pi))*lsquare[1,1]

    @property
    def Gamma3(self):
        """
        Decay rate of N3.
        """
        l         = self.h
        ldag      = np.conjugate(np.transpose(l))
        lcon      = np.conjugate(l)
        M         = self.DM
        lsquare   = np.dot(ldag,l)

        return (M[2,2]/(8*np.pi))*lsquare[2,2]

    @property
    def isPerturbative(self):
        """
        Check for perturbative nature of Yukawas
        """
        return fast_isPerturbative(self.h)
        y = self.h
        #limit of perturbativity for y
        limit = np.power(4*np.pi,0.5)
        #check if any element of column 1, 2 or 3 is larger than limit
        col1               = (y[0,0] < limit)*(y[1,0] < limit)*(y[2,0] < limit)
        col2               = (y[0,1] < limit)*(y[1,1] < limit)*(y[2,1] < limit)
        col3               = (y[0,2] < limit)*(y[1,2] < limit)*(y[2,2] < limit)
        return col1*col2*col3


if __name__ == "__main__":
    L=ULSBase()
    print(L)
    L=ULSBase(debug=True)
    print(L)



# Copyright (c) 2014, Warren Weckesser
# All rights reserved.
# See the LICENSE file for license information.

import numpy as np
from scipy.integrate import odeint


# Tuple holding the major and minor versions of numpy.
_NUMPY_VERSION = tuple(int(d) for d in np.__version__.split('.')[:2])

if _NUMPY_VERSION >= (1, 7):
    _astype_kwargs = {'copy': False}
else:
    _astype_kwargs = {}


def _check_args(kwargs):
    """
    Check for arguments that are not supported with array differential
    equations.
    """
    _unsupported_odeint_args = ['col_deriv', 'ml', 'mu']
    bad_args = [arg for arg in kwargs if arg in _unsupported_odeint_args]
    if len(bad_args) > 0:
        raise ValueError("The odeint argument %r is not supported by "
                         "odeintw for array differential equations." %
                         (bad_args[0],))


def _complex_to_real_jac(jac):
    """
    Convert a complex array to a real array with doubled dimensions.

    Each element a+i*b of `jac` is expanded to the 2x2 real block:
        [[a, -b],
         [b,  a]].

    For example, the 2x2 complex matrix::

        [1+2j  3+4j]
        [5+6j  7+8j]

    becomes the real matrix::

        [1 -2  3 -4]
        [2  1  4  3]
        [5 -6  7 -8]
        [6  5  8  7]
    """
    real_jac = np.empty((2*jac.shape[0], 2*jac.shape[1]))
    real_jac[0::2, 0::2] = jac.real
    real_jac[1::2, 1::2] = jac.real
    real_jac[0::2, 1::2] = -jac.imag
    real_jac[1::2, 0::2] = jac.imag
    return real_jac


def _transform_banded_jac(bjac):
    """
    Convert a real matrix of the form (for example)

        [0 0 A B]        [0 0 0 B]
        [0 0 C D]        [0 0 A D]
        [E F G H]   to   [0 F C H]
        [I J K L]        [E J G L]
                         [I 0 K 0]

    That is, every other column is shifted up one.
    """
    # Shift every other column.
    newjac = np.zeros((bjac.shape[0] + 1, bjac.shape[1]))
    newjac[1:, ::2] = bjac[:, ::2]
    newjac[:-1, 1::2] = bjac[:, 1::2]
    return newjac


def odeintw(func, y0, t, **kwargs):
    """
    An odeint-like function for complex array-valued differential equations.

    The function `scipy.integrate.odeint` is a wrapper of the LSODA function
    for solving ordinary differential equations.  It is designed to handle
    a system of first order differential equations expressed as a vector
    function.  `odeint` does not handle equations with complex dependent
    variables, or array equations.

    `odeintw` is a wrapper of `odeint` that adds the ability to handle
    complex and array differential equations.  See the docstring of odeint
    for an explanation of its arguments.

    Unlike odeint, all arguments after the first three position arguments
    `func` (the system definition function), `y0` (the initial condition)
    and `t` (the time values) must be given as keyword arguments.

    Initial conditions
    ------------------
    The initial condition `y0` given to `odeintw` determines the type of the
    solution that is generated.  The data type and shape of the value
    returned by `func` must match those of the initial condition.

    If the solution is to be complex, the initial condition must be complex.
    To solve a complex differential equation with real initial conditions,
    give complex initial conditions with zero imaginary parts.

    Similarly, the shape of the solution of a matrix differential equation
    is determined by the shape of the initial condition.  For example, if
    the initial condition has shape (2,3), then `func` must return a numpy
    array (or something array-like) that has shape (2,3).

    Special handling of Jacobian arguments
    --------------------------------------
    The argument `Dfun` may be used with array equations.  If `shp` is the
    shape of the array, then the shape of the Jacobian array returned by
    `Dfun` must be ``shp + shp``.  For example, if the array is 2-d `F`,
        jac[m, n, i, j] holds dF[m,n]/da[i,j]

    `Dfun` may also be used with complex equations.  However, if the
    functions computed by `func` are not complex differentiable, the
    Jacobian function should not be used.  To use the Jacobian argument in
    this case, the equations should be rewritten as a system of real
    equations for the real and imaginary parts of the solution.  For
    example, the conjugation operation is not complex differentiable, so
    to use an explicit Jacobian for the complex scalar equation
        dz/dt = z.conj(),
    the equation must be written as
        dx/dt = x
        dy/dt = -y
    Then the Jacobian of the real system is [[1, 0], [0, -1]].
    If `Dfun` is not given as an argument, the system may be left as a
    complex differential equation.

    In the case of arrays with dimension 2 or more, the odeint arguments
    `col_deriv`, `ml` and `mu` can not be used.

    Complex array equations are handled, but to use `Dfun`, the same
    requirement on the complex differentiability of the components
    holds.
    """

    full_output = kwargs.get('full_output', False)
    Dfun = kwargs.pop('Dfun', None)

    y0 = np.atleast_1d(y0)
    shape = y0.shape

    if y0.ndim == 1:
        func1 = func
        jacfunc1 = Dfun
    else:
        # y0 has dimension greater than 1.
        _check_args(kwargs)

        # Flatten y0, and create a wrapper for func that can be used
        # by odeint.
        y0 = y0.ravel()

        def vecfunc(y, t, *args):
            a = y.reshape(shape)
            dadt = func(a, t, *args)
            return np.asarray(dadt).ravel()

        func1 = vecfunc

        if Dfun is not None:

            def jacfunc(y, t, *args):
                a = y.reshape(shape)
                jac = Dfun(a, t, *args)
                return np.asarray(jac).reshape(y0.size, y0.size)

            jacfunc1 = jacfunc
        else:
            jacfunc1 = None

    if not np.iscomplexobj(y0):
        y0 = y0.astype(np.float64, **_astype_kwargs)
        func2 = func1
        jacfunc2 = jacfunc1
    else:
        # y0 is complex.
        col_deriv = kwargs.pop('col_deriv', False)
        ml = kwargs.pop('ml', None)
        mu = kwargs.pop('mu', None)
        kwargs['ml'] = None if ml is None else 1 + 2*ml
        kwargs['mu'] = None if mu is None else 1 + 2*mu

        # Cast y0 to np.complex128.
        y0 = y0.astype(np.complex128, **_astype_kwargs)

        # realfunc is a wrapper of the user's function that can be
        # used by odeint.
        def realfunc(y, t, *args):
            z = y.view(np.complex128)
            dzdt = func1(z, t, *args)
            # func might return a python list, so convert its return
            # value to an array with type np.complex128, and then return
            # a np.float64 view of that array.
            dydt = np.asarray(dzdt, dtype=np.complex128).view(np.float64)
            return dydt

        func2 = realfunc

        if jacfunc1 is not None:

            def jacfuncz(y, t, *args):
                z = y.view(np.complex128)
                jac = jacfunc1(z, t, *args)
                if col_deriv:
                    # If col_deriv is True, transpose the result returned
                    # by jacfunc1, and continue as if col_deriv was False.
                    jac = jac.T
                # Convert jac to real_jac, an array in which each complex value
                # a+i*b in jac is expanded to the 2x2 array [[a, -b], [b, a]].
                real_jac = _complex_to_real_jac(jac)
                if ml is not None or mu is not None:
                    # Banded; shift every other column up one.
                    real_jac = _transform_banded_jac(real_jac)
                return real_jac

            jacfunc2 = jacfuncz
        else:
            jacfunc2 = None

    # Call scipy.integrate.odeint with our wrapper.
    result = odeint(func2, y0.view(np.float64), t, Dfun=jacfunc2, **kwargs)

    if full_output:
        sol, infodict = result
    else:
        sol = result

    if np.iscomplexobj(y0):
        # Restore the complex view.
        sol = sol.view(np.complex128)

    # Restore the shape.
    sol = sol.reshape(-1, *shape)

    if full_output:
        result = (sol, infodict)
    else:
        result = sol
    return result



from ulysses.tools import *

from ulysses.ulsbase import ULSBase

from ulysses.etab1DME                 import EtaB_1DME
from ulysses.etab2DME                 import EtaB_2DME
from ulysses.etab3DME                 import EtaB_3DME
from ulysses.etab1BE                  import EtaB_1BE
from ulysses.etab2BE                  import EtaB_2BE
from ulysses.etab3DMEscattering       import EtaB_3DME_Scattering
from ulysses.etab3DMEscatteringRHtaur import EtaB_3DS_Scattering_RHtaur

from ulysses.etab2resonant            import EtaB_2Resonant # Buggy

testpars = {
        'delta' :270,
        'a21'   :0,
        'a31'   :0,
        't23'   :48.7,
        't12'   :33.63,
        't13'   : 8.52,
        'x1'    :45,
        'y1'    :45,
        'x2'    :45,
        'y2'    :45,
        'x3'    :45,
        'y3'    :45,
        'm'     :-0.60206,
        'M1'    :11,
        'M2'    :12,
        'M3'    :15
        }

BMpoint3 = {
        'delta' :31.713030 ,
        'a21'   :130.953483,
        'a31'   :649.655874,
        't12'   :33.630000 ,
        't23'   :46.633046 ,
        't13'   :8.520000  ,
        'x1'    :-72.335979,
        'y1'    :170.549206,
        'x2'    :86.969063 ,
        'y2'    :2.223559  ,
        'x3'    :-1.862141 ,
        'y3'    :178.312158,
        'm'     :-0.942835 ,
        'M1'    :6.500000  ,
        'M2'    :7.200000  ,
        'M3'    :7.900000
        }





















# non-resonant leptogenesis with two decaying sterile neutrino using the density matrix equations. Equations from 1112.4528
import ulysses
import numpy as np
from odeintw import odeintw

from numba import jit
@jit
def fast_RHS(y0, eps1tt,eps1mm,eps1ee,eps1tm,eps1te,eps1me,eps2tt,eps2mm,eps2ee,eps2tm,eps2te,eps2me,C,d1,d2,w1,w2,n1eq,n2eq,W):
    N1, N2, Ntt, Nmm, Nee, Ntm, Nte, Nme = y0
    widtht,widthm = W
    c1t,c1m,c1e,c2t,c2m,c2e = C

    c1tc          = np.conjugate(c1t)
    c1mc          = np.conjugate(c1m)
    c1ec          = np.conjugate(c1e)

    c2tc          = np.conjugate(c2t)
    c2mc          = np.conjugate(c2m)
    c2ec          = np.conjugate(c2e)

    #define the different RHSs for each equation
    rhs1           =      -d1*(N1-n1eq)
    rhs2           =      -d2*(N2-n2eq)
    rhs3           = (eps1tt * d1 * (N1-n1eq) + eps2tt * d2 * (N2-n2eq)
             - 0.5 * w1 * (2 * c1t * c1tc * Ntt + c1m * c1tc * Ntm + c1e * c1tc * Nte
             + np.conjugate(c1m * c1tc * Ntm + c1e * c1tc * Nte))
             - 0.5 * w2 * (2 * c2t * c2tc * Ntt + c2m * c2tc * Ntm + c2e * c2tc * Nte
             + np.conjugate(c2m * c2tc * Ntm + c2e * c2tc * Nte)))

    rhs4           = (eps1mm * d1 * (N1-n1eq) + eps2mm * d2 * (N2-n2eq)
         - 0.5 * w1 * (2 * c1m * c1mc * Nmm + c1m * c1tc * Ntm + c1e * c1mc * Nme
         + np.conjugate(c1m * c1tc * Ntm + c1e * c1mc * Nme))
         - 0.5 * w2 * (2 * c2m * c2mc * Nmm + c2m * c2tc * Ntm + c2e * c2mc * Nme
         + np.conjugate(c2m * c2tc * Ntm + c2e * c2mc * Nme)))

    rhs5           = (eps1ee * d1 * (N1-n1eq) + eps2ee * d2 * (N2-n2eq)
         - 0.5 * w1 * (2 * c1e * c1ec * Nee + c1e * c1mc * Nme + c1e * c1tc * Nte
         + np.conjugate(c1e * c1mc * Nme + c1e * c1tc * Nte))
         - 0.5 * w2 * (2 * c2e * c2ec * Nee + c2e * c2mc * Nme + c2e * c2tc * Nte
         + np.conjugate(c2e * c2mc * Nme + c2e * c2tc * Nte)))

    rhs6           = (eps1tm * d1 * (N1-n1eq) + eps2tm * d2 * (N2-n2eq)
         - 0.5 * w1 * (c1t * c1mc * Nmm + c1e * c1mc * Nte + c1m * c1mc * Ntm + c1mc * c1t * Ntt
         + c1t * c1tc * Ntm + c1t * c1ec * np.conjugate(Nme))
         - 0.5 * w2 * (c2t * c2mc * Nmm + c2e * c2mc * Nte + c2m * c2mc * Ntm + c2mc * c2t * Ntt
         + c2t * c2tc * Ntm + c2t * c2ec * np.conjugate(Nme))
         - widtht * Ntm - widthm * Ntm)

    rhs7           = (eps1te * d1 * (N1-n1eq) + eps2te * d2 * (N2-n2eq)
         - 0.5 * w1 * (c1t * c1ec * Nee + c1e * c1ec * Nte + c1m * c1ec * Ntm
         + c1t * c1ec * Ntt + c1t * c1mc * Nme + c1t * c1tc * Nte)
         - 0.5 * w2 * (c2t * c2ec * Nee + c2e * c2ec * Nte + c2m * c2ec * Ntm
         + c2t * c2ec * Ntt + c2t * c2mc * Nme + c2t * c2tc * Nte)
         - widtht * Nte)

    rhs8           = (eps1me * d1 * (N1-n1eq) + eps2me * d2 * (N2-n2eq)
         - 0.5 * w1 * (c1m * c1ec * Nee + c1e * c1ec * Nme + c1m * c1ec * Nmm
         + c1t * c1ec * np.conjugate(Ntm) + c1m * c1mc * Nme + c1m * c1tc * Nte)
         - 0.5 * w2 * (c2m * c2ec * Nee + c2e * c2ec * Nme + c2m * c2ec * Nmm
         + c2t * c2ec * np.conjugate(Ntm) + c2m * c2mc * Nme + c2m * c2tc * Nte)
         - widthm * Nme)

    RHStemp = [rhs1, rhs2, rhs3, rhs4, rhs5, rhs6, rhs7, rhs8]
    return RHStemp

class EtaB_2DME(ulysses.ULSBase):
    """
    Density matrix equation (DME) with two decaying steriles. See arxiv:1112.4528.
    """

    def RHS(self, y0, zzz, ETA, C, K, W):
        eps1tt,eps1mm,eps1ee,eps1tm,eps1te,eps1me,eps2tt,eps2mm,eps2ee,eps2tm,eps2te,eps2me = ETA
        k1term,k2term = K

        if zzz != self._currz or zzz == self.zmin:
            self._d1      = np.real(self.DS(k1term, zzz))
            self._w1      = self.scat(zzz)*np.real(self.W1(k1term, zzz))
            self._d2      = np.real(self.D2(k2term, zzz))
            self._w2      = np.real(self.W2(k2term, zzz))
            self._n1eq    = self.N1Eq(zzz)
            self._n2eq    = self.N2Eq(zzz)
            self._currz=zzz

        return fast_RHS(y0, eps1tt,eps1mm,eps1ee,eps1tm,eps1te,eps1me,eps2tt,eps2mm,eps2ee,eps2tm,eps2te,eps2me, C,self._d1,self._d2,self._w1,self._w2,self._n1eq,self._n2eq,W)

    @property
    def EtaB(self):

        #Define fixed quantities for BEs
        _ETA = [
            np.real(self.epsilon1ab(2,2)),
            np.real(self.epsilon1ab(1,1)),
            np.real(self.epsilon1ab(0,0)),
                    self.epsilon1ab(2,1) ,
                    self.epsilon1ab(2,0) ,
                    self.epsilon1ab(1,0),
            np.real(self.epsilon2ab(2,2)),
            np.real(self.epsilon2ab(1,1)),
            np.real(self.epsilon2ab(0,0)),
                    self.epsilon2ab(2,1) ,
                    self.epsilon2ab(2,0) ,
                    self.epsilon2ab(1,0),
            ]
        _C = [  self.c1a(2), self.c1a(1), self.c1a(0),
                self.c2a(2), self.c2a(1), self.c2a(0)]
        _K = [np.real(self.k1), np.real(self.k2)]
        _W = [ 485e-10*self.MP/self.M1, 1.7e-10*self.MP/self.M1]

        y0      = np.array([0+0j,0+0j,0+0j,0+0j,0+0j,0+0j,0+0j,0+0j], dtype=np.complex128)

        ys, _      = odeintw(self.RHS, y0, self.zs, args = tuple([_ETA, _C , _K, _W]), full_output=1)
        nb      = np.real(self.sphalfact*(ys[-1,2]+ys[-1,3]+ys[-1,4]))

        self.ys = np.real(ys[:, [2,3,4]])

        return np.real(nb)
